"""Add Sentiment to your dataset
"""

# Running a function across each subcluster
import numpy as np
import csv
from urllib.request import urlopen
from relevanceai.constants.errors import MissingPackageError
from relevanceai.operations.base import BaseOps


class SentimentOps(BaseOps):
    def __init__(self, model_name: str = "cardiffnlp/twitter-roberta-base-sentiment"):
        """
        Sentiment Ops.

        Parameters
        -------------

        model_name: str
            The name of the model

        """
        self.model_name = model_name

    def preprocess(self, text: str):
        new_text = []
        for t in text.split(" "):
            t = "@user" if t.startswith("@") and len(t) > 1 else t
            t = "http" if t.startswith("http") else t
            new_text.append(t)
        return " ".join(new_text)

    @property
    def classifier(self):
        if not hasattr(self, "_classifier"):
            import transformers

            self._classifier = transformers.pipeline(
                "sentiment-analysis",
                return_all_scores=True,
                model="cardiffnlp/twitter-roberta-base-sentiment",
            )
        return self._classifier

    def _get_model(self):
        try:
            import transformers
        except ModuleNotFoundError:
            print(
                "Need to install transformers by running `pip install -q transformers`."
            )
        self.classifier = transformers.pipeline(
            "sentiment-analysis",
            return_all_scores=True,
            model="cardiffnlp/twitter-roberta-base-sentiment",
        )

    def _get_label_mapping(self, task: str):
        # Note: this is specific to the current model
        labels = []
        mapping_link = f"https://raw.githubusercontent.com/cardiffnlp/tweeteval/main/datasets/{task}/mapping.txt"
        with urlopen(mapping_link) as f:
            html = f.read().decode("utf-8").split("\n")
            csvreader = csv.reader(html, delimiter="\t")
        labels = [row[1] for row in csvreader if len(row) > 1]
        return labels

    @property
    def label_mapping(self):
        return {"LABEL_0": "negative", "LABEL_1": "neutral", "LABEL_2": "positive"}

    def analyze_sentiment(
        self,
        text,
        highlight: bool = False,
        positive_sentiment_name: str = "positive",
        max_number_of_shap_documents: int = 5,
    ):
        labels = self.classifier([text])
        ind_max = np.argmax([l["score"] for l in labels[0]])
        sentiment = labels[0][ind_max]["label"]
        max_score = labels[0][ind_max]["score"]
        if sentiment == "neutral":
            overall_sentiment = 0
        else:
            overall_sentiment = (
                max_score if sentiment == positive_sentiment_name else -max_score
            )
        if not highlight:
            return {
                "sentiment": sentiment,
                "score": max_score,
                "overall_sentiment": overall_sentiment,
            }
        shap_documents = self.get_shap_values(
            text,
            sentiment_ind=ind_max,
            max_number_of_shap_documents=max_number_of_shap_documents,
        )
        return {
            "sentiment": sentiment,
            "score": max_score,
            "overall_sentiment": overall_sentiment,
            "highlight_chunk_": shap_documents,
        }

    @property
    def explainer(self):
        if hasattr(self, "_explainer"):
            return self._explainer
        else:
            try:
                import shap
            except ModuleNotFoundError:
                raise MissingPackageError("shap")
            self._explainer = shap.Explainer(self.classifier)
            return self._explainer

    def get_shap_values(
        self, text: str, sentiment_ind: int = 2, max_number_of_shap_documents: int = 5
    ):
        """Get SHAP values"""
        shap_values = self.explainer([text])
        cohorts = {"": shap_values}
        cohort_labels = list(cohorts.keys())
        cohort_exps = list(cohorts.values())
        features = cohort_exps[0].data
        feature_names = cohort_exps[0].feature_names
        values = np.array([cohort_exps[i].values for i in range(len(cohort_exps))])
        shap_docs = [
            {"text": v, "score": f}
            for f, v in zip(
                [x[sentiment_ind] for x in values[0][0].tolist()], feature_names[0]
            )
        ]
        return sorted(shap_docs, key=lambda x: x["score"], reverse=True)[
            :max_number_of_shap_documents
        ]

    # def analyze_sentiment(self, text, highlight:bool= True):
    #     try:
    #         from scipy.special import softmax
    #     except ModuleNotFoundError:
    #         print("Need to install scipy")
    #     if not hasattr(self, "model"):
    #         self._get_model()
    #     text = self.preprocess(text)
    #     encoded_input = self.tokenizer(text, return_tensors="pt")
    #     output = self.model(**encoded_input)
    #     scores = output[0][0].detach().numpy()
    #     scores = softmax(scores)
    #     ranking = np.argsort(scores)
    #     ranking = ranking[::-1]
    #     sentiment = self.label_mapping[ranking[0]]
    #     score = np.round(float(scores[ranking[0]]), 4)
    #     return {
    #         "sentiment": sentiment,
    #         "score": np.round(float(scores[ranking[0]]), 4),
    #         "overall_sentiment_score": score if sentiment == "positive" else -score,
    #     }



from relevanceai import Client
from relevanceai import ClusterOps
from relevanceai.dataset import Dataset


class TestBaseOps:
    def test_init(self, test_client: Client):
        credentials = test_client.credentials
        ops = ClusterOps.init(credentials=credentials)
        assert True

    def test_from_credentials(self, test_client: Client):
        credentials = test_client.credentials
        ops = ClusterOps.init(credentials=credentials)
        assert ops.credentials.token == credentials.token

    def test_from_token(self, test_token: str):
        ops = ClusterOps.from_token(token=test_token)
        assert ops.credentials.token == test_token

    def test_from_client(self, test_client: Client):
        ops = ClusterOps.from_client(client=test_client)
        assert ops.credentials.token == test_client.credentials.token

    def test_from_dataset(self, test_clustered_df: Dataset):
        ops = ClusterOps.from_dataset(
            dataset=test_clustered_df,
            alias="kmeans-10",
            vector_fields=["sample_1_vector_"],
        )
        closest = ops.list_closest()
        assert closest



import tempfile
import shutil
import os
import shutil
from relevanceai.constants import CONFIG_PATH, Config


class ConfigMixin:
    def _create_temp_config(self):
        tmp = tempfile.NamedTemporaryFile(delete=False)
        shutil.copy2(CONFIG_PATH, tmp.name)
        return tmp.name

    @property
    def config_path(self):
        # Create a temporary config path
        if not hasattr(self, "_config_path"):
            # generate
            self._config_path = self._create_temp_config()
        return self._config_path

    @property
    def config(self):
        if not hasattr(self, "_config"):
            self._config = Config(self.config_path)
        return self._config

    @config.setter
    def config(self, config: Config):
        self._config = config



from functools import wraps

from relevanceai.utils.logger import FileLogger


def log(fn: str = "logs.txt", verbose: bool = False, log_to_file=True):
    def _log(f):

        with FileLogger(fn=fn, verbose=verbose, log_to_file=log_to_file):

            @wraps(f)
            def wrapper(*args, **kwds):
                return f(*args, **kwds)

        return wrapper

    return _log



from relevanceai.dataset import Dataset


def test_reduce_dimensions(test_dataset: Dataset):
    model = "pca"
    alias = "pca"
    n_components = 3
    vector_field = "sample_1_vector_"
    test_dataset.reduce_dims(
        model=model,
        n_components=n_components,
        vector_fields=[vector_field],
        alias=alias,
    )

    dr_vector_name = f"_dr_.{alias}.{vector_field}"
    assert dr_vector_name in test_dataset.schema



from typing import Dict, List, Optional, Union

from relevanceai.client.helpers import Credentials
from relevanceai.dataset.read import Read
from relevanceai.dataset.write import Write
from relevanceai.dataset.io import IO
from relevanceai.dataset.series import Series
from relevanceai.operations import Operations
from relevanceai.utils.decorators.analytics import track
from relevanceai.constants import (
    GLOBAL_DATASETS,
    SEARCH_APP_LINK,
    PROJECTOR_APP_LINK,
)


class Dataset(IO, Read, Write, Operations):
    @track
    def __init__(
        self,
        credentials: Credentials,
        dataset_id: str,
        fields: Optional[list] = None,
        image_fields: Optional[List[str]] = None,
        audio_fields: Optional[List[str]] = None,
        highlight_fields: Optional[Dict[str, List]] = None,
        text_fields: Optional[List[str]] = None,
        **kwargs,
    ):
        self.credentials = credentials
        self.fields = [] if fields is None else fields
        self.dataset_id = dataset_id
        self.image_fields = [] if image_fields is None else image_fields
        self.audio_fields = [] if audio_fields is None else audio_fields
        self.highlight_fields = {} if highlight_fields is None else highlight_fields
        self.text_fields = [] if text_fields is None else text_fields

        super().__init__(
            credentials=credentials,
            dataset_id=self.dataset_id,
            fields=fields,
            image_fields=image_fields,
            audio_fields=audio_fields,
            highlight_fields=highlight_fields,
            text_fields=text_fields,
            **kwargs,
        )
        # add global datasets
        if self.dataset_id in GLOBAL_DATASETS:
            # avoid re-inserting if it already exists
            if self.dataset_id not in self.datasets.list()["datasets"]:
                from relevanceai.utils.datasets import mock_documents
                from relevanceai.utils.decorators.analytics import fire_and_forget

                @fire_and_forget
                def add_mock_dataset():
                    self.upsert_documents(mock_documents(100))

                add_mock_dataset()
        self.is_empty()

    def is_empty(self):
        """Check if a dataset is empty."""
        try:
            if self.dataset_id not in self.datasets.list()["datasets"]:
                try:
                    print("⚠️ Your dataset has no documents. Make sure to insert some!")
                except:
                    pass
        except:
            pass

    @track
    def __getitem__(self, field: Union[List[str], str]):
        """
        Returns a Series Object that selects a particular field within a dataset
        Parameters
        ----------
        field: Union[List, str]
            The particular field within the dataset
        Returns
        -------
        Tuple
            (N, C)
        Example
        ---------------
        .. code-block::

            from relevanceai import Client
            client = Client()
            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)
            field = "sample_field"
            series = df[field]
        """
        if isinstance(field, str):
            return Series(
                credentials=self.credentials,
                dataset_id=self.dataset_id,
                field=field,
                image_fields=self.image_fields,
                audio_fields=self.audio_fields,
                highlight_fields=self.highlight_fields,
                text_fields=self.text_fields,
            )
        elif isinstance(field, list):
            return Dataset(
                credentials=self.credentials,
                dataset_id=self.dataset_id,
                fields=field,
                image_fields=self.image_fields,
                audio_fields=self.audio_fields,
                highlight_fields=self.highlight_fields,
                text_fields=self.text_fields,
            )
        else:
            raise TypeError("Field needs to be a list or a string.")

    @track
    def launch_search_app(self):
        """
        Launches the link to the search application to start building

        .. code-block::

            ds.launch_search_app()

        """
        return SEARCH_APP_LINK.format(self.dataset_id)

    @track
    def launch_projector_app(self):
        """
        Launches the link to the projector application to start building

        .. code-block::

            ds.launch_projector_app()

        """
        return PROJECTOR_APP_LINK.format(self.dataset_id)



from relevanceai.operations.vector.vectorize import VectorizeOps
from relevanceai.operations.vector.search import SearchOps
from relevanceai.operations.vector.base import Base2Vec
from relevanceai.operations.vector.vectorizer import Vectorizer



from typing import (
    Any,
    Set,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np

from relevanceai._api import APIClient
from relevanceai.client.helpers import Credentials
from relevanceai.constants.errors import MissingPackageError
from relevanceai.dataset import Dataset
from relevanceai.operations import BaseOps
from relevanceai.utils.decorators import beta, track, deprecated
from relevanceai.operations import BaseOps
from relevanceai.constants import (
    Warning,
    Messages,
    CLUSTER_APP_LINK,
    ModelNotSupportedError,
)
from relevanceai.operations.cluster.models.summarizer import TransformersLMSummarizer

from relevanceai.operations.cluster.utils import ClusterUtils
from doc_utils import DocUtils


class ClusterOps(ClusterUtils, BaseOps, DocUtils):
    """
    You can load ClusterOps instances in 2 ways.

    .. code-block::

        # State the vector fields and alias in the ClusterOps object
        cluster_ops = client.ClusterOps(
            alias="kmeans-16",
            dataset_id="sample_dataset_id",
            vector_fields=['sample_vector_']
        )

        cluster_ops.list_closest()

        # State the vector fields and alias in the operational call
        cluster_ops = client.ClusterOps(alias="kmeans-16")
        cluster_ops.list_closest(
            dataset="sample_dataset_id",
            vector_fields=["documentation_vector_]
        )

    """

    dataset_id: str
    cluster_field: str

    def __init__(
        self,
        credentials: Credentials,
        model: Any = None,
        alias: str = None,
        n_clusters: Optional[int] = None,
        cluster_config: Optional[Dict[str, Any]] = None,
        outlier_value: int = -1,
        outlier_label: str = "outlier",
        verbose: bool = True,
        vector_fields: Optional[list] = None,
        **kwargs,
    ):
        """
        ClusterOps object

        Parameters
        -------------

        model: Any
            The string of clustering algorithm, class of clustering algorithm or custom clustering class.
            If custom, the model must contain the method for fit_predict and must output a numpy array for labels

        alias: str = None
            An alias to be given for clustering.
            If no alias is provided, alias becomes the cluster model name hypen n_clusters e.g. (kmeans-5, birch-8)

        n_clusters: Optional[int] = None
            Number of clusters to find, should the argument be required of the algorithm.
            If None, n_clusters defaults to 8 for sklearn models ONLY.

        cluster_config: Optional[Dict[str, Any]] = None
            Hyperparameters for the clustering model.
            See the documentation for supported clustering models for list of hyperparameters

        outlier_value: int = -1
            For unsupervised clustering models like OPTICS etc. outliers are given the integer label -1

        outlier_label: str = "outlier"
            When viewing the cluster app dashboard, outliers will be prefixed with outlier_label

        """

        self.cluster_config = {} if cluster_config is None else cluster_config  # type: ignore

        self.model_name = None

        if model is None:
            model = "community_detection"
            if verbose:
                print(f"No clustering model selected: defaulting to `{model}`")

        self.n_clusters = n_clusters

        if n_clusters is not None:
            self.cluster_config["n_clusters"] = n_clusters  # type: ignore

        self.model = self._get_model(model)

        if self.n_clusters is None:
            if hasattr(self.model, "n_clusters"):
                self.n_clusters = self.model.n_clusters

            elif hasattr(self.model, "k"):
                self.n_clusters = self.model.k

        if "package" not in self.__dict__:
            self.package = self._get_package(self.model)

        self.alias = self._get_alias(alias)
        self.outlier_value = outlier_value
        self.outlier_label = outlier_label

        for key, value in kwargs.items():
            if not hasattr(self, key):
                if key == "vector_fields":
                    setattr(self, "vector_field", value[0])
                setattr(self, key, value)

        super().__init__(credentials)

    def __call__(
        self,
        dataset_id: str,
        vector_fields: Optional[List[str]] = None,
        include_cluster_report: bool = True,
        **kwargs,
    ) -> None:
        return self.run(
            dataset_id=dataset_id,
            vector_fields=vector_fields,
            include_cluster_report=include_cluster_report,
            **kwargs,
        )

    @deprecated(version="1.0.0")
    def fit_predict_update(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def _get_schema(self) -> Dict:
        return self.datasets.schema(dataset_id=self.dataset_id)

    def _get_alias(self, alias: Any) -> str:
        # Auto-generates alias here
        if alias is None:
            if hasattr(self.model, "n_clusters"):
                n_clusters = (
                    self.n_clusters
                    if self.n_clusters is not None
                    else self.model.n_clusters
                )
                alias = f"{self.model_name}-{n_clusters}"

            elif hasattr(self.model, "k"):
                n_clusters = (
                    self.n_clusters if self.n_clusters is not None else self.model.k
                )
                alias = f"{self.model_name}-{n_clusters}"

            else:
                alias = self.model_name

            Warning.MISSING_ALIAS.format(alias=alias)

        print(f"The alias is `{alias.lower()}`.")
        return alias.lower()

    def _get_package(self, model):
        model_name = str(model.__class__).lower()
        if "function" in model_name:
            model_name = str(model.__name__)

        if "sklearn" in model_name:
            package = "sklearn"

        elif "faiss" in model_name:
            package = "faiss"

        elif "hdbscan" in model_name:
            package = "hdbscan"

        elif "communitydetection" in model_name:
            package = "sentence-transformers"

        else:
            package = "custom"

        return package

    def _get_model(self, model):
        if isinstance(model, str):
            model = model.lower().replace(" ", "").replace("_", "")
            self.model_name = model
        else:
            self.model_name = str(model.__class__).split(".")[-1].split("'>")[0]
            self.package = "custom"
            return model

        if isinstance(model, str):
            if model == "affinitypropagation":
                from sklearn.cluster import AffinityPropagation

                model = AffinityPropagation(**self.cluster_config)

            elif model == "agglomerativeclustering":
                from sklearn.cluster import AgglomerativeClustering

                model = AgglomerativeClustering(**self.cluster_config)

            elif model == "birch":
                from sklearn.cluster import Birch

                model = Birch(**self.cluster_config)

            elif model == "dbscan":
                from sklearn.cluster import DBSCAN

                model = DBSCAN(**self.cluster_config)

            elif model == "optics":
                from sklearn.cluster import OPTICS

                model = OPTICS(**self.cluster_config)

            elif model == "kmeans":
                from sklearn.cluster import KMeans

                model = KMeans(**self.cluster_config)

            elif model == "featureagglomeration":
                from sklearn.cluster import FeatureAgglomeration

                model = FeatureAgglomeration(**self.cluster_config)

            elif model == "meanshift":
                from sklearn.cluster import MeanShift

                model = MeanShift(**self.cluster_config)

            elif model == "minibatchkmeans":
                from sklearn.cluster import MiniBatchKMeans

                model = MiniBatchKMeans(**self.cluster_config)

            elif model == "spectralclustering":
                from sklearn.cluster import SpectralClustering

                model = SpectralClustering(**self.cluster_config)

            elif model == "spectralbiclustering":
                from sklearn.cluster import SpectralBiclustering

                model = SpectralBiclustering(**self.cluster_config)

            elif model == "spectralcoclustering":
                from sklearn.cluster import SpectralCoclustering

                model = SpectralCoclustering(**self.cluster_config)

            elif model == "hdbscan":
                from hdbscan import HDBSCAN

                model = HDBSCAN(**self.cluster_config)

            elif model in "communitydetection":
                from relevanceai.operations.cluster.algorithms import CommunityDetection

                model = CommunityDetection(**self.cluster_config)

            elif "faiss" in model:
                from faiss import Kmeans

                model = Kmeans(**self.cluster_config)
            else:
                raise ValueError(
                    """Invalid model. This should be one of ['affinitypropagation',
                    'agglomerativeclustering', 'birch', dbscan', 'optics', 'kmeans',
                    'featureagglomeration', 'meanshift', 'minibatchkmeans',
                    'spectralclustering', 'spectralbiclustering', 'spectralcoclustering',
                    'hdbscan', 'communitydetection']
                    ]"""
                )

        else:
            raise ModelNotSupportedError

        return model

    def _format_labels(self, labels: np.ndarray) -> List[str]:
        labels = labels.flatten().tolist()
        cluster_labels = [
            f"cluster-{str(label)}"
            if label != self.outlier_value
            else self.outlier_label
            for label in labels
        ]
        return cluster_labels

    def _get_centroid_documents(
        self, vectors: np.ndarray, labels: List[str], vector_field: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        centroid_documents = []

        centroids: Dict[str, Any] = {}
        for label, vector in zip(labels, vectors.tolist()):
            if label not in centroids:
                centroids[label] = []
            centroids[label].append(vector)

        for centroid, vectors in centroids.items():
            centroid_vector = np.array(vectors).mean(0).tolist()
            if vector_field:
                centroid_document = {
                    "_id": centroid,
                    vector_field: centroid_vector,
                }
            else:
                centroid_document = dict(
                    _id=centroid,
                    centroid_vector=centroid_vector,
                )
            centroid_documents.append(centroid_document)

        return centroid_documents

    def _get_document_vector_field(self):
        vector_fields = []
        metadata = self.datasets.metadata(self.dataset_id)
        most_recent_updated_vector = max(
            metadata["results"]["_vector_"].items(), key=lambda x: x[1]
        )[0]
        vector_fields.append(most_recent_updated_vector)
        return vector_fields

    def _insert_centroids(
        self,
        dataset_id: str,
        vector_fields: List[str],
        centroid_documents: List[Dict[str, Any]],
    ) -> None:
        self.services.cluster.centroids.insert(
            dataset_id=dataset_id,
            cluster_centers=centroid_documents,
            vector_fields=vector_fields,
            alias=self.alias,
        )

    def _fit_predict(
        self, documents: List[Dict[str, Any]], vector_field: str
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        vectors = np.array(
            [self.get_field(vector_field, document) for document in documents]
        )

        if self.package == "sklearn":
            labels = self.model.fit_predict(vectors)

        elif self.package == "hdbscan":
            self.model.fit(vectors)
            labels = self.model.labels_.reshape(-1, 1)

        elif self.package == "faiss":
            vectors = vectors.astype("float32")
            self.model.train(vectors)
            labels = self.model.assign(vectors)[1]

        elif self.package == "sentence-transformers":
            # TODO: make teh config here better
            labels = self.model(vectors)
            labels = np.array(labels)

        elif self.package == "custom":
            labels = self.model.fit_predict(vectors)
            if not isinstance(labels, np.ndarray):
                raise ValueError("Custom model must output np.ndarray for labels")

            if labels.shape[0] != vectors.shape[0]:
                raise ValueError("incorrect shape")

        labels = self._format_labels(labels)

        if self.n_clusters is None:
            n_clusters = len(set(labels)) - 1
            print(f"Found {n_clusters} clusters using {self.model_name}")

        self.set_field_across_documents(
            field=self.cluster_field,
            values=labels,
            docs=documents,
        )

        centroid_documents = self._get_centroid_documents(
            vectors, labels, vector_field=vector_field
        )

        return centroid_documents, documents

    def _print_app_link(self):
        link = CLUSTER_APP_LINK.format(self.dataset_id)
        print(Messages.BUILD_HERE + link)

    @track
    def run(
        self,
        dataset_id: str,
        vector_fields: Optional[List[str]] = None,
        filters: Optional[list] = None,
        show_progress_bar: bool = True,
        verbose: bool = True,
        include_cluster_report: bool = True,
        report_name: str = "cluster-report",
    ) -> None:
        """
        Run clustering on a dataset

        Parameters
        --------------

        dataset_id: Optional[Union[str, Any]]
            The dataset ID
        vector_fields: Optional[List[str]]
            List of vector fields
        show_progress_bar: bool
            If True, the progress bar can be shown

        """
        filters = [] if filters is None else filters
        if not isinstance(dataset_id, str):
            if hasattr(dataset_id, "dataset_id"):
                dataset_id = dataset_id.dataset_id  # type: ignore
        self.dataset_id = dataset_id

        if vector_fields is None:
            vector_fields = self._get_document_vector_field()
            print(f"No vector_field given: defaulting to {vector_fields}")

        vector_field = vector_fields[0]
        self.cluster_field = f"_cluster_.{vector_fields[0]}.{self.alias}"

        # get all documents
        print("Retrieving all documents...")
        from relevanceai.utils.filter_helper import create_filter

        filters = create_filter(vector_field, filter_type="exists")
        documents = self._get_all_documents(
            dataset_id=dataset_id,
            select_fields=vector_fields,
            show_progress_bar=show_progress_bar,
            include_vector=True,
            filters=filters,
        )

        # fit model, predict and label all documents
        print("Predicting on all documents...")
        centroid_documents, labelled_documents = self._fit_predict(
            documents=documents,
            vector_field=vector_field,
        )

        print("Updating cluster labels...")
        results = self._update_documents(
            dataset_id=dataset_id,
            documents=labelled_documents,
            show_progress_bar=show_progress_bar,
        )

        print("Inserting Centroids...")
        self._insert_centroids(
            dataset_id=dataset_id,
            vector_fields=vector_fields,
            centroid_documents=centroid_documents,
        )
        if include_cluster_report:
            print("Generating evaluation report for your clusters…")
            from relevanceai.reports.cluster.report import ClusterReport

            centroids = self.get_field_across_documents(
                vector_field, centroid_documents, missing_treatment="raise_error"
            )

            X = self.get_field_across_documents(
                vector_field, documents, missing_treatment=self.outlier_value
            )

            cluster_labels = self.get_field_across_documents(
                self.cluster_field, documents, missing_treatment=-1
            )

            if len(cluster_labels) != len(X):
                raise ValueError("Damn son. How you like them apples.")

            report = ClusterReport(
                X=X,
                cluster_labels=cluster_labels,
                model=self.model,
                outlier_label=-1,
                centroids=centroids,
                verbose=True,
                include_typecheck=False,
            )

            try:
                response = self.reports.clusters.create(
                    name=report_name, report=self.json_encoder(report.internal_report)
                )

                if verbose:
                    print(
                        f"📊 You can now access your report at https://cloud.relevance.ai/report/cluster/{self.region}/{response['_id']}"
                    )
            except Exception as e:
                print("Error creating cluster report! " + str(e))

        # link back to dashboard
        if verbose:
            self._print_app_link()

    def cluster_report(self, X, cluster_labels, centroids):
        from relevanceai.reports.cluster.report import ClusterReport

        cluster_labels = self.fit_predict(X)
        centroids = ClusterReport.calculate_centroids(X, cluster_labels)
        report = ...
        response = self.store_cluster_report(
            report_name="kmeans", report=report.internal_report
        )

    @track
    def closest(
        self,
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 3,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        cluster_properties_filter: Optional[Dict] = {},
        verbose: bool = True,
    ):
        """
        List of documents closest from the center.

        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        vector_field: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        centroid_vector_fields: list
            Vector fields stored
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        cluster_properties_filter: dict
            Filter if clusters with certain characteristics should be hidden in results

        """
        dataset_id = self.dataset_id if dataset_id is None else dataset_id
        vector_field = self.vector_field if vector_field is None else vector_field
        alias = self.alias if alias is None else alias

        return self.services.cluster.centroids.list_closest_to_center(
            dataset_id=dataset_id,
            vector_fields=[vector_field],
            alias=alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            select_fields=select_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            cluster_properties_filter=cluster_properties_filter,
            verbose=verbose,
        )

    @staticmethod
    def get_cluster_summary(
        summarizer,
        docs: Dict,
        summarize_fields: List[str],
        max_length: int = 100,
        first_sentence_only: bool = True,
    ):
        def _clean_sentence(s):
            s = (
                s.replace(". .", ".")
                .replace(" .", ".")
                .replace("\n", "")
                .replace("..", ".")
                .strip()
            )
            if s[-1] != ".":
                s += "."
            return s

        cluster_summary = {}
        for cluster, results in docs["results"].items():
            summary = []
            for f in summarize_fields:
                summary_fields = [
                    _clean_sentence(d[f])
                    for d in results["results"]
                    if d.get(f) and d[f] not in [" ", "."]
                ]
                summary_output = summarizer(
                    " ".join(summary_fields), max_length=max_length
                )[0]["summary_text"]
                summary = summary_output.replace(" .", ".").strip()
            if first_sentence_only:
                cluster_summary[cluster] = summary.split(".")[0]
            else:
                cluster_summary[cluster] = summary
        return cluster_summary

    @beta
    @track
    def summarize_closest(
        self,
        summarize_fields: List[str],
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 3,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        cluster_properties_filter: Optional[Dict] = {},
        model_name: str = "sshleifer/distilbart-cnn-6-6",
        tokenizer: Optional[str] = None,
        max_length: int = 100,
        deployable_id: Optional[str] = None,
        first_sentence_only: bool = True,
        **kwargs,
    ):
        """
        List of documents closest from the center.

        Parameters
        ----------
        summarize_fields: list
            Fields to perform summarization, empty array/list means all fields
        dataset_id: string
            Unique name of dataset
        vector_field: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        centroid_vector_fields: list
            Vector fields stored
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        cluster_properties_filter: dict
            Filter if clusters with certain characteristics should be hidden in results
        model_name: str
            Huggingface Model to use for summarization.
            Pick from
            https://huggingface.co/models?pipeline_tag=summarization&sort=downloadshttps://huggingface.co/models?pipeline_tag=summarization
        tokenizer: str
            Tokenizer to use for summarization, allows you to bring your own tokenizer,
            else will instantiate pre-trained from selected model

        """
        dataset_id = self.dataset_id if dataset_id is None else dataset_id
        vector_field = self.vector_field if vector_field is None else vector_field
        alias = self.alias if alias is None else alias

        if not tokenizer:
            tokenizer = model_name

        summarizer = TransformersLMSummarizer(model_name, tokenizer, **kwargs)

        center_docs = self.list_closest(
            select_fields=summarize_fields,
            dataset_id=dataset_id,
            vector_field=vector_field,
            alias=alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            cluster_properties_filter=cluster_properties_filter,
        )

        cluster_summary = self.get_cluster_summary(
            summarizer,
            docs=center_docs,
            summarize_fields=summarize_fields,
            max_length=max_length,
            first_sentence_only=first_sentence_only,
        )

        if deployable_id is not None:
            if dataset_id is None:
                if not hasattr(self, "dataset_id"):
                    raise ValueError("You need a dataset ID to update.")
                else:
                    dataset_id = self.dataset_id
            configuration = self.deployables.get(deployable_id=deployable_id)
            configuration["cluster-labels"] = cluster_summary
            self.deployables.update(
                deployable_id=deployable_id,
                dataset_id=dataset_id,
                configuration=configuration,
            )
        return {"results": cluster_summary}

    @track
    def furthest(
        self,
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 3,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        cluster_properties_filter: Optional[Dict] = {},
    ):
        """
        List documents furthest from the center.

        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        vector_fields: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        """
        dataset_id = self.dataset_id if dataset_id is None else dataset_id
        vector_field = self.vector_field if vector_field is None else vector_field
        alias = self.alias if alias is None else alias

        return self.services.cluster.centroids.list_furthest_from_center(
            dataset_id=dataset_id,
            vector_fields=[vector_field],
            alias=alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            select_fields=select_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            cluster_properties_filter=cluster_properties_filter,
        )

    @beta
    @track
    def summarize_furthest(
        self,
        summarize_fields: List[str],
        dataset_id: Optional[str] = None,
        vector_field: Optional[str] = None,
        alias: Optional[str] = None,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 3,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        # facets: List = [],
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        cluster_properties_filter: Optional[Dict] = {},
        model_name: str = "sshleifer/distilbart-cnn-6-6",
        tokenizer: Optional[str] = None,
        **kwargs,
    ):
        """
        List of documents furthest from the center.

        Parameters
        ----------
        summarize_fields: list
            Fields to perform summarization, empty array/list means all fields
        dataset_id: string
            Unique name of dataset
        vector_field: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        centroid_vector_fields: list
            Vector fields stored
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        cluster_properties_filter: dict
            Filter if clusters with certain characteristics should be hidden in results
        model_name: str
            Huggingface Model to use for summarization.
            Pick from
            https://huggingface.co/models?pipeline_tag=summarization&sort=downloadshttps://huggingface.co/models?pipeline_tag=summarization
        tokenizer: str
            Tokenizer to use for summarization, allows you to bring your own tokenizer,
            else will instantiate pre-trained from selected model

        """
        dataset_id = self.dataset_id if dataset_id is None else dataset_id
        vector_field = self.vector_field if vector_field is None else vector_field
        alias = self.alias if alias is None else alias

        if not tokenizer:
            tokenizer = model_name
        summarizer = TransformersLMSummarizer(model_name, tokenizer)

        center_docs = self.list_furthest(
            select_fields=summarize_fields,
            dataset_id=dataset_id,
            vector_field=vector_field,
            alias=alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
            cluster_properties_filter=cluster_properties_filter,
        )

        cluster_summary = self.get_cluster_summary(
            summarizer, docs=center_docs, summarize_fields=summarize_fields
        )

        return {"results": cluster_summary}

    # Convenience functions
    list_closest = closest
    list_furthest = furthest

    def _retrieve_dataset_id(self, dataset: Optional[Union[str, Dataset]]) -> str:
        """Helper method to get multiple dataset values"""
        if isinstance(dataset, Dataset):
            dataset_id: str = dataset.dataset_id
        elif isinstance(dataset, str):
            dataset_id = dataset
        elif dataset is None:
            if hasattr(self, "dataset_id"):
                # let's not surprise users
                print(
                    f"No dataset supplied - using last stored one '{self.dataset_id}'."
                )
                dataset_id = str(self.dataset_id)
            else:
                raise ValueError("Please supply dataset.")
        return dataset_id

    @track
    def aggregate(
        self,
        vector_fields: List[str] = None,
        metrics: Optional[list] = None,
        sort: Optional[list] = None,
        groupby: Optional[list] = None,
        filters: Optional[list] = None,
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
        dataset: Optional[Union[str, Dataset]] = None,
    ):
        """
        Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.
        It can only can be used after a vector field has been clustered. \n
        Aggregation/Groupby of a collection using an aggregation query. The aggregation query is a json body that follows the schema of:

        .. code-block::

            {
                "groupby" : [
                    {"name": <alias>, "field": <field in the collection>, "agg": "category"},
                    {"name": <alias>, "field": <another groupby field in the collection>, "agg": "numeric"}
                ],
                "metrics" : [
                    {"name": <alias>, "field": <numeric field in the collection>, "agg": "avg"}
                    {"name": <alias>, "field": <another numeric field in the collection>, "agg": "max"}
                ]
            }

        For example, one can use the following aggregations to group score based on region and player name.

        .. code-block::

            {
                "groupby" : [
                    {"name": "region", "field": "player_region", "agg": "category"},
                    {"name": "player_name", "field": "name", "agg": "category"}
                ],
                "metrics" : [
                    {"name": "average_score", "field": "final_score", "agg": "avg"},
                    {"name": "max_score", "field": "final_score", "agg": "max"},
                    {'name':'total_score','field':"final_score", 'agg':'sum'},
                    {'name':'average_deaths','field':"final_deaths", 'agg':'avg'},
                    {'name':'highest_deaths','field':"final_deaths", 'agg':'max'},
                ]
            }
        "groupby" is the fields you want to split the data into. These are the available groupby types:
            - category : groupby a field that is a category
            - numeric: groupby a field that is a numeric
        "metrics" is the fields and metrics you want to calculate in each of those, every aggregation includes a frequency metric. These are the available metric types:
            - "avg", "max", "min", "sum", "cardinality"
        The response returned has the following in descending order. \n
        If you want to return documents, specify a "group_size" parameter and a "select_fields" parameter if you want to limit the specific fields chosen. This looks as such:
            .. code-block::
                {
                    'groupby':[
                        {'name':'Manufacturer','field':'manufacturer','agg':'category',
                        'group_size': 10, 'select_fields': ["name"]},
                    ],
                    'metrics':[
                        {'name':'Price Average','field':'price','agg':'avg'},
                    ],
                }
                # ouptut example:
                {"title": {"title": "books", "frequency": 200, "documents": [{...}, {...}]}, {"title": "books", "frequency": 100, "documents": [{...}, {...}]}}
        For array-aggregations, you can add "agg": "array" into the aggregation query.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten
        alias: string
            Alias used to name a vector field. Belongs in field_{alias} vector
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten

        Example
        ---------
        .. code-block::
            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")
            from sklearn.cluster import KMeans
            model = KMeans(n_clusters=2)
            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
            cluster_ops.run(df, vector_fields=["sample_vector_"])
            clusterer.aggregate(
                "sample_dataset_id",
                groupby=[{
                    "field": "title",
                    "agg": "wordcloud",
                }],
                vector_fields=['sample_vector_']
            )
        """
        metrics = [] if metrics is None else metrics
        groupby = [] if groupby is None else groupby
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort

        return self.services.cluster.aggregate(
            dataset_id=self._retrieve_dataset_id(dataset),
            vector_fields=self.vector_fields if not vector_fields else vector_fields,
            groupby=groupby,
            metrics=metrics,
            sort=sort,
            filters=filters,
            alias=self.alias,
            page_size=page_size,
            page=page,
            asc=asc,
            flatten=flatten,
        )

    @track
    def merge(
        self,
        cluster_labels: List,
        alias: Optional[str] = None,
        show_progress_bar: bool = True,
    ):
        """
        Parameters
        ----------
        cluster_labels : Tuple[int]
            a tuple of integers representing the cluster ids you would like to merge

        alias: str
            the alias of the clustering you like to merge labels within

        show_progress_bar: bool
            whether or not to show the progress bar

        Example
        -------

        .. code-block::

            dataset.cluster(
                model="kmeans",
                n_clusters=3,
                vector_fields=["sample_1_vector_"],
            )

            ops = ClusterOps.from_dataset(
                dataset=dataset,
                alias="kmeans-3",
                vector_fields=["sample_1_vector_"],
            )

            ops.merge(cluster_labels=(0, 1), alias="kmeans-3")

        """

        if alias is None:
            alias = "communitydetection"
            print("No alias given, assuming `communitydetection`")

        centroid_documents = self.services.cluster.centroids.list(
            dataset_id=self.dataset_id,
            vector_fields=[self.vector_field],
            alias=alias,
            include_vector=True,
        )["results"]

        relevant_centroids = [
            centroid[self.vector_field]
            for centroid in centroid_documents
            if any(f"-{cluster}" in centroid["_id"] for cluster in cluster_labels)
        ]
        new_centroid = np.array(relevant_centroids).mean(0).tolist()
        if isinstance(cluster_labels[0], int):
            new_centroid_doc = {
                "_id": f"cluster-{cluster_labels[0]}",
                self.vector_field: new_centroid,
            }
        elif isinstance(cluster_labels[0], str):
            if isinstance(cluster_labels[0], int):
                new_centroid_doc = {
                    "_id": cluster_labels,
                    self.vector_field: new_centroid,
                }

        class Merge(DocUtils):
            def __init__(self, clusters, vector_field, alias):
                if isinstance(clusters[0], str):
                    self.clusters = [cluster for cluster in sorted(clusters)]
                    self.min_cluster = clusters[0]
                else:
                    self.clusters = [
                        f"cluster-{cluster}" for cluster in sorted(clusters)
                    ]
                    self.min_cluster = f"cluster-{min(clusters)}"

                self.vector_field = vector_field
                self.alias = alias

            def __call__(self, documents):
                for document in documents:
                    for cluster in self.clusters[1:]:
                        if (
                            self.get_field(
                                f"_cluster_.{self.vector_field}.{self.alias}", document
                            )
                            == cluster
                        ):
                            self.set_field(
                                f"_cluster_.{self.vector_field}.{self.alias}",
                                document,
                                self.min_cluster,
                            )
                return documents

        merge = Merge(cluster_labels, self.vector_field, alias)
        self.pull_update_push(
            dataset_id=self.dataset_id,
            update_function=merge,
            show_progress_bar=show_progress_bar,
        )

        self.services.cluster.centroids.update(
            dataset_id=self.dataset_id,
            vector_fields=[self.vector_field],
            alias=alias,
            cluster_centers=[new_centroid_doc],
        )

        cluster: int

        for cluster in cluster_labels[1:]:
            if isinstance(cluster, str):
                centroid_id = cluster
            else:
                centroid_id = f"cluster-{cluster}"
            self.services.cluster.centroids.delete(
                dataset_id=self.dataset_id,
                centroid_id=centroid_id,
                alias=self.alias,
                vector_fields=[self.vector_field],
            )

    def create_centroids(self):
        """
        Calculate centroids

        Example
        --------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            cluster_ops = ds.ClusterOps(
                alias="kmeans-40",
                vector_fields=['text_roberta-large_vector_']
            )
            centroids = cluster_ops.create_centroids()

        """
        # Get an array of the different vectors
        if len(self.vector_fields) > 1:
            raise NotImplementedError(
                "Do not currently support multiple vector fields for centroid creation."
            )
        centroid_vectors = {}

        def calculate_centroid(vectors):
            X = np.array(vectors)
            return X.mean(axis=0)

        centroid_vectors = self._operate_across_clusters(
            field=self.vector_fields[0], func=calculate_centroid
        )

        self._insert_centroids(
            dataset_id=self.dataset_id,
            vector_fields=[self.vector_fields[0]],
            centroid_documents=centroid_vectors,
        )
        return centroid_vectors



import traceback

from tqdm import tqdm

from collections import defaultdict
from typing import Optional

from relevanceai.dataset.write.write import Write

from relevanceai.utils.decorators.analytics import track
from relevanceai.utils.logger import FileLogger


class CommunityDetection(Write):
    pass


#     @track
#     def community_detection(
#         self,
#         field: str,
#         model=None,
#         retrieval_kwargs: Optional[dict] = None,
#         encode_kwargs: Optional[dict] = None,
#         threshold: float = 0.75,
#         min_community_size: int = 3,
#         init_max_size: int = 1000,
#         update_chunksize: int = 100,
#         alias: str = "community-detection",
#         log_file: str = "logs.txt",
#         cluster_field: str = "_cluster_",
#     ):
#         """
#         Performs community detection on a text field.

#         Parameters
#         ----------
#         field: str
#             The field over which to find communities. Must be of type "text"
#             or "vector".

#         model
#             A model for computing sentence embeddings.

#         retrieval_kwargs: Optional[dict]
#             Keyword arguments for `get_documents` call. See respective
#             details for argument details.

#         encode_kwargs: Optional[dict]
#             Keyword arguments for the provide model's `encode` call. See
#             respective method for argument details.

#         threshold: float
#             A lower limit of similarity that determines whether two embeddings
#             are similar or not.

#         min_community_size: int
#             The minimum size of a community. Only communities that are larger
#             than this value are returned, and the first element of each
#             community is treated as the central point.

#         init_max_size: int
#             The maximum size of a community. If the corpus is larger than this
#             value, that is set to the maximum size.

#         Example
#         -------

#         .. code-block::

#             from relevanceai import Client

#             client = Client()

#             ds = client.Dataset("sample_dataset_id")

#             communities = ds.community_detection("sample_text_field")

#         """
#         if field in self.schema:
#             if not (self.schema[field] == "text" or field.endswith("_vector_")):
#                 raise ValueError("The field must be a 'text' type or a vector")
#         else:
#             raise ValueError(f"{field} does not exist in the dataset")

#         try:
#             with FileLogger(log_file):
# from sentence_transformers.util import community_detection
#         except ModuleNotFoundError:
#             raise ModuleNotFoundError(
#                 "community_detection function not found. "
#                 "Please install sentence-transformers with `python -m "
#                 "pip install -U sentence-transformers` to install "
#                 "community_detection."
#             )

#         field_type = "text" if self.schema[field] == "text" else "vector"

#         retrieval_kwargs = {} if retrieval_kwargs is None else retrieval_kwargs
#         encode_kwargs = {} if encode_kwargs is None else encode_kwargs

#         if model is None and field_type == "text":
#             from sentence_transformers import SentenceTransformer

#             model = SentenceTransformer("all-MiniLM-L6-v2")
#             # encode defaults:
#             #  batch_size: int = 32
#             #  show_progress_bar: bool = None
#             #  output_value: str = 'sentence_embedding'
#             #  convert_to_numpy: bool = True
#             #  convert_to_Tensor: bool = False
#             #  device: str = None
#             #  normalize_embeddings: bool = False

#         print("Retrieving documents...")
#         documents = self.get_all_documents(
#             select_fields=[field],
#             filters=[
#                 {
#                     "field": field,
#                     "filter_type": "exists",
#                     "condition": "==",
#                     "condition_value": " ",
#                 }
#             ],
#             **{
#                 key: value
#                 for key, value in retrieval_kwargs.items()
#                 if key != "select_fields"
#             },
#         )
#         print("Documents retrieved.")

#         # lists (i.e. vectors) are unhashable therefore handled by converting
#         # them into tuples.
#         if field_type == "vector":
#             for document in documents:
#                 try:
#                     value = tuple(self.get_field(field, document))
#                     self.set_field(field, document, value)
#                 except KeyError:
#                     # If a document is missing a vector, ignore
#                     continue

#         # Keep track of the fields. Since two documents could have the same
#         # field, use a list to keep track of multiples
#         element_ids = defaultdict(list)

#         # remove duplicates
#         elements = set()
#         for document in documents:
#             try:
#                 element = self.get_field(field, document)
#             except Exception as e:
#                 # It could be that a document does not have a field or a
#                 # a vector that other documents in the Dataset has. In that
#                 # case, ignore.
#                 traceback.print_exc()
#                 continue
#             elements.add(element)
#             element_ids[element].append(document["_id"])

#         # Storing a mapping like this is fine because when the values are
#         # made a list below, they will be a list in the same order as inserted
#         # in the dictionary.
#         element_map = {}
#         ids_map = {}
#         for i, element in enumerate(elements):
#             element_map[i] = element
#             ids_map[i] = element_ids[element]

#         if field_type == "text":
#             print("Encoding the corpus...")
#             embeddings = model.encode(
#                 sentences=list(element_map.values()),
#                 **{
#                     key: value
#                     for key, value in encode_kwargs.items()
#                     if key != "sentences"
#                 },
#             )
#             print("Encoding complete.")
#         else:
#             from numpy import array

#             embeddings = array(list(element_map.values()))

#         print("Community detection started...")
#         init_max_size = min(init_max_size, embeddings.shape[0])
#         clusters = community_detection(
#             embeddings=embeddings,
#             threshold=threshold,
#             min_community_size=min_community_size,
#             init_max_size=init_max_size,
#         )
#         print(f"Detected {len(clusters)} communities.")

#         # TODO: add centroids for community detection

#         print("Updating documents...")
#         community_documents = []
#         centroids: List[dict] = []
#         for i, cluster in enumerate(tqdm(clusters)):
#             ids = []
#             centroid = cluster[0]
#             centroids.append(
#                 {
#                     "_id": ids_map[centroid][0],  # choose the first ID as centroid
#                     "centroid_vector_": embeddings[centroid].tolist(),
#                 }
#             )
#             for member in cluster:
#                 ids.extend(ids_map[member])
#             # During initial construction update_where did not accept dict
#             # values as valid updates.
#             # self.datasets.documents.update_where(
#             #    self.dataset_id,
#             #    update={
#             #        "_cluster_": {field: {"community-detection": f"community-{i+1}"}}
#             #    },
#             #    filters=[
#             #        {
#             #            "field": "ids",
#             #            "filter_type": "ids",
#             #            "condition": "==",
#             #            "condition_value": ids,
#             #        }
#             #    ],
#             # )
#             for id in ids:
#                 community_documents.append(
#                     {
#                         "_id": id,
#                         cluster_field: {field: {alias: f"cluster-{i+1}"}},
#                     }
#                 )

#             # During initial construction update_where did not accept dict
#             # values as valid updates.
#             with FileLogger(log_file):
#                 result = self.datasets.documents.update_where(
#                     self.dataset_id,
#                     update={"_cluster_": {field: {alias: f"cluster-{i+1}"}}},
#                     filters=[
#                         {
#                             "field": "ids",
#                             "filter_type": "ids",
#                             "condition": "==",
#                             "condition_value": ids,
#                         }
#                     ],
#                 )
#                 print(result)

#         print(
#             "Build your clustering app here: "
#             f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
#         )
#         # Return a ClusterOps object

#         from relevanceai.workflows.cluster_ops.ops import ClusterOps

#         cluster_ops: ClusterOps = ClusterOps(
#             # model=model,
#             alias=alias,
#             dataset_id=self.dataset_id,
#             vector_fields=[field],
#             cluster_field=cluster_field,
#             project=self.project,
#             api_key=self.api_key,
#             firebase_uid=self.firebase_uid,
#             verbose=False,
#         )
#         print("Creating centroids...")
#         with FileLogger(log_file):
#             result = cluster_ops.insert_centroid_documents(centroids)
#         print("✅ Uploaded centroids.")
#         return cluster_ops



from relevanceai.operations.text_finetuning.supervised_finetuning_ops import (
    SupervisedTripleLossFinetuneOps,
)
from relevanceai.operations.text_finetuning.unsupervised_finetuning_ops import GPLOps



# -*- coding: utf-8 -*-
"""
Pandas like dataset API
"""
import warnings
import numpy as np

from typing import List, Optional
from tqdm.auto import tqdm

from relevanceai.utils.decorators.analytics import track
from relevanceai.utils.decorators.version import beta
from relevanceai.constants.warning import Warning

# from relevanceai.operations.cluster.models.community_detection import CommunityDetection

# class AutoClusterOps(CommunityDetection):
#     @track
#     def cluster(self, model, alias, vector_fields, **kwargs):
#         """
#         Performs KMeans Clustering on over a vector field within the dataset.

#         .. warning::
#             Deprecated in v0.33 in favour of df.auto_cluster.

#         Parameters
#         ------------
#         model : Class
#             The clustering model to use
#         vector_fields : str
#             The vector fields over which to cluster

#         Example
#         ------------
#         .. code-block::

#             from relevanceai import Client
#             from relevanceai ClusterOps
#             from relevanceai.ops.clusterops.kmeans_clusterer import KMeansModel

#             client = Client()

#             dataset_id = "sample_dataset_id"
#             df = client.Dataset(dataset_id)

#             vector_field = "vector_field_"
#             n_clusters = 10

#             model = KMeansModel(k=n_clusters)

#             df.cluster(model=model, alias=f"kmeans-{n_clusters}", vector_fields=[vector_field])
#         """
#         from relevanceai.operations.cluster import ClusterOps

#         clusterer = ClusterOps(
#             credentials=self.credentials,
#             model=model,
#             alias=alias,
#         )
#         clusterer.fit(dataset=self, vector_fields=vector_fields)
#         return clusterer

#     def cluster_keyphrases(
#         self,
#         vector_fields: List[str],
#         text_fields: List[str],
#         cluster_alias: str,
#         cluster_field: str = "_cluster_",
#         num_clusters: int = 100,
#         most_common: int = 10,
#         preprocess_hooks: Optional[List[callable]] = None,
#         algorithm: str = "rake",
#         n: int = 2,
#     ):
#         """
#         Simple implementation of the cluster word cloud.

#         Parameters
#         ------------
#         vector_fields: list
#             The list of vector fields
#         text_fields: list
#             The list of text fields
#         cluster_alias: str
#             The alias of the cluster
#         cluster_field: str
#             The cluster field to try things on
#         num_clusters: int
#             The number of clusters
#         preprocess_hooks: list
#             The preprocess hooks
#         algorithm: str
#             The algorithm to use
#         n: int
#             The number of words

#         """
#         preprocess_hooks = [] if preprocess_hooks is None else preprocess_hooks

#         vector_fields_str = ".".join(sorted(vector_fields))
#         field = f"{cluster_field}.{vector_fields_str}.{cluster_alias}"
#         all_clusters = self.facets([field], page_size=num_clusters)
#         cluster_counters = {}
#         if "results" in all_clusters:
#             all_clusters = all_clusters["results"]
#         # TODO: Switch to multiprocessing
#         for c in tqdm(all_clusters[field]):
#             cluster_value = c[field]
#             top_words = self.keyphrases(
#                 text_fields=text_fields,
#                 n=n,
#                 filters=[
#                     {
#                         "field": field,
#                         "filter_type": "contains",
#                         "condition": "==",
#                         "condition_value": cluster_value,
#                     }
#                 ],
#                 most_common=most_common,
#                 preprocess_hooks=preprocess_hooks,
#                 algorithm=algorithm,
#             )
#             cluster_counters[cluster_value] = top_words
#         return cluster_counters

#     # TODO: Add keyphrases to auto cluster
#     def auto_cluster_keyphrases(
#         vector_fields: List[str],
#         text_fields: List[str],
#         cluster_alias: str,
#         deployable_id: str,
#         n: int = 2,
#         cluster_field: str = "_cluster_",
#         num_clusters: int = 100,
#         preprocess_hooks: Optional[List[callable]] = None,
#     ):
#         """
#         # TODO:
#         """
#         pass

#     def _add_cluster_word_cloud_to_config(self, data, cluster_value, top_words):
#         # TODO: Add this to wordcloud deployable
#         # hacky way I implemented to add top words to config
#         data["configuration"]["cluster-labels"][cluster_value] = ", ".join(
#             [k for k in top_words if k != "machine learning"]
#         )
#         data["configuration"]["cluster-descriptions"][cluster_value] = str(top_words)

#     @track
#     def auto_cluster(
#         self,
#         alias: str,
#         vector_fields: List[str],
#         model=None,
#         chunksize: int = 1024,
#         filters: Optional[list] = None,
#         parent_alias: Optional[str] = None,
#     ):
#         """
#         Automatically cluster in 1 line of code.
#         It will retrieve documents, run fitting on the documents and then
#         update the database.
#         There are only 2 supported clustering algorithms at the moment:
#         - kmeans
#         - minibatchkmeans

#         In order to choose the number of clusters, simply add a number
#         after the dash like `kmeans-8` or `minibatchkmeans-50`.

#         Under the hood, it uses scikit learn defaults or best practices.

#         This returns a ClusterOps object and is a wrapper on top of
#         `ClusterOps`.

#         Parameters
#         ----------
#         alias : str
#             The clustering model (as a str) to use and n_clusters. Delivered in a string separated by a '-'
#             Supported aliases at the moment are 'kmeans','kmeans-10', 'kmeans-X' (where X is a number), 'minibatchkmeans',
#                 'minibatchkmeans-10', 'minibatchkmeans-X' (where X is a number)
#         vector_fields : List
#             A list vector fields over which to cluster

#         Example
#         ----------

#         .. code-block::

#             from relevanceai import Client

#             client = Client()

#             dataset_id = "sample_dataset_id"
#             df = client.Dataset(dataset_id)

#             # run kmeans with default 10 clusters
#             clusterer = df.auto_cluster("kmeans", vector_fields=[vector_field])
#             clusterer.list_closest()

#             # Run k means clustering with 8 clusters
#             clusterer = df.auto_cluster("kmeans-8", vector_fields=[vector_field])

#             # Run minibatch k means clustering with 8 clusters
#             clusterer = df.auto_cluster("minibatchkmeans-8", vector_fields=[vector_field])

#             # Run minibatch k means clustering with 20 clusters
#             clusterer = df.auto_cluster("minibatchkmeans-20", vector_fields=[vector_field])

#         You can alternatively run this using kmeans.

#         .. code-block::

#             from relevanceai import Client

#             client = Client()

#             from relevanceai.package_utils.datasets import mock_documents

#             ds = client.Dataset('sample')
#             ds.upsert_documents(mock_documents(100))
#             # Run initial kmeans to get clusters
#             ds.auto_cluster('kmeans-3', vector_fields=["sample_1_vector_"])
#             # Run separate K Means to get subclusters
#             cluster_ops = ds.auto_cluster(
#                 'kmeans-2',
#                 vector_fields=["sample_1_vector_"],
#                 parent_alias="kmeans-3"
#             )

#         """
#         if model is None:
#             return self._auto_cluster_string(
#                 alias=alias,
#                 vector_fields=vector_fields,
#                 chunksize=1024,
#                 filters=filters,
#                 parent_alias=parent_alias,
#             )
#         else:
#             return self._auto_cluster_model(
#                 alias=alias,
#                 model=model,
#                 vector_fields=vector_fields,
#                 chunksize=chunksize,
#                 filters=filters,
#                 parent_alias=parent_alias,
#             )

#     def _auto_cluster_model(
#         self,
#         alias: str,
#         vector_fields: List[str],
#         chunksize: int = 1024,
#         filters: Optional[list] = None,
#         parent_alias: Optional[str] = None,
#         model=None,
#     ):
#         filters = [] if filters is None else filters

#         cluster_args = alias.split("-")
#         algorithm = cluster_args[0]
#         if len(cluster_args) > 1:
#             n_clusters = int(cluster_args[1])
#         else:
#             print("No clusters are detected, defaulting to 8")
#             n_clusters = 8
#         if n_clusters >= chunksize:
#             raise ValueError("Number of clustesr exceed chunksize.")

#         num_documents = self.get_number_of_documents(self.dataset_id)

#         if num_documents <= n_clusters:
#             warnings.warn(Warning.NCLUSTERS_GREATER_THAN_NDOCS)

#         from relevanceai.operations.cluster import ClusterOps

#         clusterer: ClusterOps = ClusterOps(
#             credentials=self.credentials,
#             model=model,
#             alias=alias,
#         )
#         if parent_alias:
#             clusterer.subcluster_predict_update(
#                 dataset=self,
#                 vector_fields=vector_fields,
#                 filters=filters,
#             )
#         else:
#             clusterer.fit(
#                 dataset_id=self.dataset_id,
#                 vector_fields=vector_fields,
#                 parent_alias=parent_alias,
#                 include_grade=True,
#                 filters=filters,
#             )
#         return clusterer

#     def _store_subcluster_metadata(
#         self, vector_fields: list, alias: str, parent_alias: str
#     ):
#         # Store metadata around subclustering
#         field = str("-".join(vector_fields)) + "." + alias
#         metadata = self.metadata
#         if "subcluster" not in metadata:
#             metadata["subclusters"] = []
#         metadata["subclusters"].append(
#             {
#                 "parent_alias": parent_alias,
#                 "alias": alias,
#                 "vector_fields": vector_fields,
#             }
#         )
#         self.upsert_metadata(metadata)

#     @track
#     def _auto_cluster_string(
#         self,
#         alias: str,
#         vector_fields: List[str],
#         chunksize: int = 1024,
#         filters: Optional[list] = None,
#         parent_alias: Optional[str] = None,
#     ):
#         """
#         Automatically cluster in 1 line of code.
#         It will retrieve documents, run fitting on the documents and then
#         update the database.
#         There are only 2 supported clustering algorithms at the moment:
#         - kmeans
#         - minibatchkmeans

#         In order to choose the number of clusters, simply add a number
#         after the dash like `kmeans-8` or `minibatchkmeans-50`.

#         Under the hood, it uses scikit learn defaults or best practices.

#         This returns a ClusterOps object and is a wrapper on top of
#         `ClusterOps`.

#         Parameters
#         ----------
#         alias : str
#             The clustering model (as a str) to use and n_clusters. Delivered in a string separated by a '-'
#             Supported aliases at the moment are 'kmeans','kmeans-10', 'kmeans-X' (where X is a number), 'minibatchkmeans',
#                 'minibatchkmeans-10', 'minibatchkmeans-X' (where X is a number)
#         vector_fields : List
#             A list vector fields over which to cluster

#         Example
#         ----------

#         .. code-block::

#             from relevanceai import Client

#             client = Client()

#             dataset_id = "sample_dataset_id"
#             df = client.Dataset(dataset_id)

#             # run kmeans with default 10 clusters
#             clusterer = df.auto_cluster("kmeans", vector_fields=[vector_field])
#             clusterer.list_closest()

#             # Run k means clustering with 8 clusters
#             clusterer = df.auto_cluster("kmeans-8", vector_fields=[vector_field])

#             # Run minibatch k means clustering with 8 clusters
#             clusterer = df.auto_cluster("minibatchkmeans-8", vector_fields=[vector_field])

#             # Run minibatch k means clustering with 20 clusters
#             clusterer = df.auto_cluster("minibatchkmeans-20", vector_fields=[vector_field])

#         You can alternatively run this using kmeans.

#         .. code-block::

#             from relevanceai import Client

#             client = Client()

#             from relevanceai.package_utils.datasets import mock_documents

#             ds = client.Dataset('sample')
#             ds.upsert_documents(mock_documents(100))
#             # Run initial kmeans to get clusters
#             ds.auto_cluster('kmeans-3', vector_fields=["sample_1_vector_"])
#             # Run separate K Means to get subclusters
#             cluster_ops = ds.auto_cluster(
#                 'kmeans-2',
#                 vector_fields=["sample_1_vector_"],
#                 parent_alias="kmeans-3"
#             )

#         """
#         filters = [] if filters is None else filters

#         cluster_args = alias.split("-")
#         algorithm = cluster_args[0]
#         if len(cluster_args) > 1:
#             n_clusters = int(cluster_args[1])
#         else:
#             print("No clusters are detected, defaulting to 8")
#             n_clusters = 8
#         if n_clusters >= chunksize:
#             raise ValueError("Number of clustesr exceed chunksize.")

#         num_documents = self.get_number_of_documents(self.dataset_id)

#         if num_documents <= n_clusters:
#             warnings.warn(
#                 "You seem to have more clusters than documents. We recommend reducing the number of clusters."
#             )

#         from relevanceai.operations.cluster import ClusterOps

#         if algorithm.lower() == "kmeans":
#             from sklearn.cluster import KMeans

#             model = KMeans(n_clusters=n_clusters)
#             clusterer: ClusterOps = ClusterOps(
#                 credentials=self.credentials,
#                 model=model,
#                 alias=alias,
#                 dataset_id=self.dataset_id,
#                 vector_fields=vector_fields,
#                 parent_alias=parent_alias,
#             )
#             if parent_alias:
#                 clusterer.subcluster_predict_update(
#                     dataset=self,
#                     vector_fields=vector_fields,
#                     filters=filters,
#                 )

#                 self._store_subcluster_metadata(
#                     vector_fields=vector_fields, alias=alias, parent_alias=parent_alias
#                 )
#             else:
#                 clusterer.fit(
#                     dataset=self,
#                     vector_fields=vector_fields,
#                     include_report=True,
#                     filters=filters,
#                 )

#         elif algorithm.lower() == "hdbscan":
#             raise ValueError(
#                 "HDBSCAN is soon to be released as an alternative clustering algorithm"
#             )
#         elif algorithm.lower() == "minibatchkmeans":
#             from sklearn.cluster import MiniBatchKMeans

#             model = MiniBatchKMeans(n_clusters=n_clusters)

#             clusterer = ClusterOps(
#                 credentials=self.credentials,
#                 model=model,
#                 alias=alias,
#                 dataset_id=self.dataset_id,
#                 vector_fields=vector_fields,
#                 parent_alias=parent_alias,
#             )

#             if parent_alias:
#                 print("subpartial fit...")
#                 clusterer.subpartialfit_predict_update(
#                     dataset=self,
#                     vector_fields=vector_fields,
#                     filters=filters,
#                 )

#                 self._store_subcluster_metadata(
#                     vector_fields=vector_fields, alias=alias, parent_alias=parent_alias
#                 )

#             else:
#                 clusterer.partial_fit_predict_update(
#                     dataset=self,
#                     vector_fields=vector_fields,
#                     chunksize=chunksize,
#                     filters=filters,
#                 )

#         elif algorithm.lower() == "communitydetection":
#             if len(vector_fields) > 1:
#                 raise ValueError(
#                     "Currently we do not support more than 1 vector field."
#                 )
#             return self.community_detection(field=vector_fields[0], alias=alias)
#         else:
#             raise ValueError("Only KMeans clustering is supported at the moment.")

#         return clusterer

#     def auto_text_cluster_dashboard(
#         self,
#         text_fields: List[str],
#         alias: str,
#         chunksize: int = 1024,
#         filters: Optional[list] = None,
#         text_encoder=None,
#     ):
#         """
#         Convenient way to vectorize and cluster text fields.

#         Parameters
#         ----------
#         text_fields: List[str]
#             A list of text fields to vectorize and cluster

#         alias: str
#             The name of the clustring application. The alias is required to
#             be of the form "{algorithm}-{n_clusters}" where:
#                 * algorithm is the clustering algorithm to be used; and
#                 * n_clusters is the number of clusters

#         chunksize: int
#             The size of the chunks

#         filters: Optional[list]
#             A list of filters to apply over the fields to vectorize

#         text_encoder:
#             A deep learning text encoder from the vectorhub library. If no
#             encoder is specified, a default encoder (USE2Vec) is loaded.

#         Returns
#         -------
#         A dictionary indicating the outcome status

#         Example
#         -------

#         .. code-block::

#             from relevanceai import Client

#             client = Client()

#             ds = client.Dataset("sample_dataset_id")

#             ds.auto_text_cluster_dashboard(text_fields=["sample_text_field"])

#         """
#         filters = [] if filters is None else filters

#         fields = []
#         for field in text_fields:
#             try:
#                 if not "text" == self.schema[field]:
#                     fields.append(field)
#             except KeyError:
#                 raise KeyError(f"'{field}' is an invalid field")
#         else:
#             if fields:
#                 raise ValueError(
#                     "The following fields are not text fields: " f"{', '.join(fields)}"
#                 )

#         results = self.vectorize(text_fields=text_fields, text_encoder=text_encoder)
#         if "added_vectors" not in results:
#             # If there were errors in vectorizing, then quit immediately and return errors
#             return results

#         new_vectors = results["added_vectors"]
#         existing_vectors = results["skipped_vectors"]

#         self.auto_cluster(
#             alias=alias,
#             vector_fields=new_vectors + existing_vectors,
#             chunksize=chunksize,
#             filters=filters,
#         )

#         print(
#             "Build your clustering app here: "
#             f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
#         )

#     def launch_cluster_app(self, configuration: dict = None):
#         """
#         Launch an app with a given configuration


#         Example
#         --------

#         .. code-block::

#             ds.launch_cluster_app()

#         Parameters
#         -----------

#         configuration: dict
#             The configuration can be found in the deployable once created.

#         """
#         if configuration is None:
#             url = f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
#             print(
#                 "Build your clustering app here: "
#                 f"https://cloud.relevance.ai/dataset/{self.dataset_id}/deploy/recent/cluster"
#             )
#             return url
#         if "configuration" in configuration:
#             configuration = configuration["configuration"]
#         results = self.deployables.create(
#             dataset_id=self.dataset_id, configuration=configuration
#         )

#         # After you have created an app
#         url = f"https://cloud.relevance.ai/dataset/{results['dataset_id']}/deploy/cluster/{self.project}/{self.api_key}/{results['deployable_id']}/{self.region}"
#         print(f"You can now access your deployable at {url}.")
#         return url
